/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "anal.h"
#include "tree.h"
#include "stack.h"

static gp2c_stack s_node, s_value;
static gp2c_stack s_opcode, s_data;

static long *opcode;
static GEN *data;
static long offset=-1;

const char *opname[]={"_||_", "_&&_", "_==_", "_!=_", "_>=_", "_>_", "_<=_", "_<_", "_-_","_+_","_<<_", "_>>_", "_%_", "_\\/_", "_\\_", "_/_", "_*_","_^_","__","_=_","_--","_++","_-=_", "_+=_", "_<<=_", "_>>=_", "_%=_", "_\\/=_", "_\\=_", "_/=_", "_*=_","+_","-_","!_","_!","_'","_~","%","length","",""};

void
init_compiler(void)
{
  long i;
  struct node_loc l;
  char *noarg="<No argument>";
  stack_init(&s_node,sizeof(*tree),(void **)&tree);
  stack_init(&s_value,sizeof(*value),(void **)&value);
  l.start=noarg; l.end=noarg+strlen(noarg);
  for (i=0;i<OPnboperator;i++)
    newentry(opname[i]);
  stack_init(&s_opcode,sizeof(*opcode),(void **)&opcode);
  stack_init(&s_data,sizeof(*data),(void **)&data);
}

long startnode(void)
{ 
  return s_node.n-1;
}

void
freenode(void)
{
  long i;
  s_node.n=0;
  for (i=OPnboperator; i<s_value.n; i++)
    switch(value[i].type)
    {
    case CSTstr: case CSTquote:
      gpfree((void*)value[i].val.str);
      break;
    case CSTreal: case CSTgen:
      gunclone(value[i].val.gen);
      break;
    }
  s_value.n=OPnboperator;
}

int newnode(Ffunc f, int x, int y, struct node_loc *loc)
{
  int n=stack_new(&s_node);
  tree[n].f=f;
  tree[n].x=x;
  tree[n].y=y;
  tree[n].start=loc->start;
  tree[n].end=loc->end;
  return n;
}

int newopcall(OPerator op, int x, int y, struct node_loc *loc)
{
  if (y==-1)
    return newnode(Ffunction,op,x,loc);
  else
    return newnode(Ffunction,op,newnode(Flistarg,x,y,loc),loc);
}

int newstringvalue(const char *s)
{
  int n=stack_new(&s_value);
  value[n].type=CSTstr;
  value[n].val.str=s;
  return n++;
}

int newquotevalue(const char *s)
{
  int n=stack_new(&s_value);
  value[n].type=CSTquote;
  value[n].val.str=s;
  return n++;
}

int newrealvalue(GEN g)
{
  int n=stack_new(&s_value);
  value[n].type=CSTreal;
  value[n].val.gen=gclone(g);
  return n++;
}

int newgenvalue(GEN g)
{
  int n=stack_new(&s_value);
  value[n].type=CSTgen;
  value[n].val.gen=gclone(g);
  return n++;
}

int newgennode(GEN g, struct node_loc *loc)
{
  if (typ(g)==t_INT)
  {
    long s;
    if (signe(g)==0) return newnode(Fsmall,0,-1,loc);
    if ((s=itos_or_0(g))) return newnode(Fsmall,s,-1,loc);
    return newnode(Fconst,newgenvalue(g),-1,loc);
  }
  return newnode(Fconst,newrealvalue(g),-1,loc);
}

int newentry(const char *s)
{
  return newstringvalue(s);
}

int newmember(const char *s)
{
  char *p;
  int n;
  p=gpmalloc(3+strlen(s));
  sprintf(p,"_.%s",s);
  n=newentry(p);
  gpfree((char *)s);
  return n;
}

void
getcodepos(long *sop, long *sdat, long *off)
{
  *sop=s_opcode.n;
  *sdat=s_data.n;
  *off=offset;
  offset=s_data.n-1;
}

GEN
getclosure(long sop, long sdat, long off)
{
  long lop =s_opcode.n+1-sop;
  long ldat=s_data.n+1-sdat;
  GEN cl=cgetg(3,t_VEC);
  long i;
  gel(cl,1) = cgetg(lop,  t_VECSMALL);
  gel(cl,2) = cgetg(ldat, t_VEC);
  for(i=1;i<lop;i++)
    mael(cl, 1, i) = opcode[i+sop-1];
  s_opcode.n=sop;
  for(i=1;i<ldat;i++)
  {
    gmael(cl, 2, i) = gcopy(data[i+sdat-1]);
    gunclone(data[i+sdat-1]);
  }
  s_data.n=sdat;
  offset=off;
  return cl;
}

void
op_push(op_code o, long x)
{
  int n=stack_new(&s_opcode);
  int m=stack_new(&s_opcode);
  opcode[n]=o;
  opcode[m]=x;
}

long
data_push(GEN x)
{
  int n=stack_new(&s_data);
  data[n] = gclone(x);
  return n-offset;
}

